#!/usr/bin/python3
# -*- coding: utf-8 -*-


import os, sys, json, hashlib
import io
from PIL import Image, ImageChops, ImageDraw, ImageFont

from eimagedata import *

class EimageDataError (ValueError):
    pass

class EimageError (Exception):
    pass

class EimageProcessor:
    
    DEFAULT_ADOPT_MAX_STEP = 16
    DEFAULT_RESIZE_ALGO_MAX_HEIGHT  = 100
    DEFAULT_TITLE_CONF = {
        "float": "bottom",
        "height": 0,
        "grid": (3, 3),
        "align": ("left", "center"),
        "padding": (3, 0),
        "font": {"name": "cour.ttf", "size": 14}
    }
    DEFAULT_TITLE_TRANSFORM= {"a": 1, "b": 0, "c": 0, "d": 1, "x": 0, "y": 0}

    ImageAsPIL = 1
    ImageAsJPEG = 2


    def __init__(self):
        self.ADOPT_MAX_STEP = self.DEFAULT_ADOPT_MAX_STEP
        self.RESIZE_ALGO_MAX_HEIGHT = self.DEFAULT_RESIZE_ALGO_MAX_HEIGHT
        self.titlefont = None 
        if 'linux' in sys.platform:
            self.titlefont = next((f for f in ["/usr/share/fonts/truetype/ttf-dejavu/DejaVuSans.ttf", 
                    "/usr/share/fonts/truetype/dejavu/DejaVuSans.ttf"] if os.path.isfile(f)), None)
        elif 'win' in sys.platform:
            self.titlefont =  "arial.ttf"
        if not self.titlefont:
            raise EimageError('Couldn`t determine subtitle font to use')
        self.subtitles = dict(EIMAGE_SUBTITLES)
        self.presets = dict(EIMAGE_PRESETS)
        self.lang = tuple(EIMAGE_LANGUAGES)

    def ihash(self,idata,dig):
        ih = dict()
        if idata:
            ih['size'] = len(idata)
            if not dig or 'md5' in dig: ih['md5'] = hashlib.md5(idata).hexdigest()
            if not dig or 'sha1' in dig: ih['sha1'] = hashlib.sha1(idata).hexdigest()
            if not dig or 'sha256' in dig: ih['sha256'] = hashlib.sha256(idata).hexdigest()
        return ih
    
    def iinfo(self,img):
        ii = dict()
        if img:
            ii['size'] = img.size
            ii['bbox'] = img.getbbox()
            ii['extrema'] = img.getextrema()
            ii['format'] = img.format
            ii['mode'] = img.mode
            ii['extra'] = img.info
        return ii

    def transform(self, rect, transform, box):
        return [
            (rect[0] * transform['a'] + transform['x']) * box[0],
            (rect[1] * transform['d'] + transform['y']) * box[1],
            (rect[2] * transform['a'] + transform['x']) * box[0],
            (rect[3] * transform['d'] + transform['y']) * box[1]
        ]
        
    def recrop(self, ims, i_rect, i_margin, i_sizes, nodata = False):
        """
            i_rect: [left, top, right, bottom],  # float[4] relative to image width (height)
            
            i_sizes: [                        # int[6] may be int[4] so adaptives to be omitted

                absolute-width,                # if no other sizes exist - to be an absolute width and height 
                absolute-height,            # without keeping geometry (pass edge clip)

                keep-ratio-width,            # if only one exists - force other to keep existing image ratio
                keep-ratio-height,            # if both exists - select max result dim and keep ratio too (pass edge clip)

                adaptive-width,                # adaptive algorythm uses to preserve this dims moving margins 
                adaptive-height ],            # to avoid edge clip and black bars
            
            i_margin: [                        # float[8] polynom coeff. i.e. left-margin = (lmtw * width + lmth * height)
                                            # may be float[4] so height coeff to be omitted
                                            # may be negative
                left-margin-to-width,        # 
                left-margin-to-height,        #
                top-margin-to-width,        #
                top-margin-to-height,        #
                right-margin-to-width,        #
                right-margin-to-height,        #
                bottom-margin-to-width,        #
                bottom-margin-to-height,]    #
                
        """
        
        (ims_w, ims_h) = ims.size

        nocrop = noresize = False
        
        if not i_sizes:
            (fw, fh, frw, frh, faw, fah) = (None, None, None, None, None, None)
            noresize = True
        elif len(i_sizes) == 4:
            (fw, fh, frw, frh, faw, fah) = (i_sizes[0], i_sizes[1], i_sizes[2],
                                            i_sizes[3], None, None)
        elif len(i_sizes) == 6:
            (fw, fh, frw, frh, faw, fah) = (i_sizes[0], i_sizes[1], i_sizes[2],
                                            i_sizes[3], i_sizes[4], i_sizes[5])
        else:
            raise EimageDataError('Invalid sizes')

        (fw, fh, frw, frh, faw, fah) = (
            fw if (fw is not None and fw > 0) else None,
            fh if (fh is not None and fh > 0) else None,
            frw if (frw is not None and frw > 0) else None,
            frh if (frh is not None and frh > 0) else None,
            faw,
            fah
        )
        
        if not i_rect or i_rect == [0,0,0,0]:
            nocrop = True
        else:    
            (rl, rt, rr, rb) = (
                int(i_rect[0] * ims_w),
                int(i_rect[1] * ims_h),
                int(i_rect[2] * ims_w),
                int(i_rect[3] * ims_h)
            )
            (rl, rt, rr, rb) = (
                0 if (rl is None or rl < 1 or rl > ims_w or rl >= rr) else rl,
                0 if (rt is None or rt < 1 or rt > ims_h or rt >= rb) else rt,
                0 if (rr is None or rr < 1 or rr > ims_w or rr <= rl) else rr,
                0 if (rb is None or rb < 1 or rb > ims_h or rb <= rt) else rb
            )
        
        if not nocrop:
            if not i_margin:
                (lmcw, lmch, tmcw, tmch, rmcw, rmch, bmcw, bmch) = (0.0, 0.0, 0.0, 0.0, 
                                                                    0.0, 0.0, 0.0, 0.0,)
            elif len(i_margin) == 4:
                (lmcw, lmch, tmcw, tmch, rmcw, rmch, bmcw, bmch) = (i_margin[0], 0.0,
                                                                    i_margin[1], 0.0,
                                                                    i_margin[2], 0.0,
                                                                    i_margin[3], 0.0)
            elif len(i_margin) == 8:
                (lmcw, lmch, tmcw, tmch, rmcw, rmch, bmcw, bmch) = (
                    i_margin[0],
                    i_margin[1],
                    i_margin[2],
                    i_margin[3],
                    i_margin[4],
                    i_margin[5],
                    i_margin[6],
                    i_margin[7]
                )
            else:
                raise EimageDataError('Invalid margin')
            
            (lmcw, lmch, tmcw, tmch, rmcw, rmch, bmcw, bmch) = (
                0 if (lmcw is None) else lmcw,
                0 if (lmch is None) else lmch,
                0 if (tmcw is None) else tmcw,
                0 if (tmch is None) else tmch,
                0 if (rmcw is None) else rmcw,
                0 if (rmch is None) else rmch,
                0 if (bmcw is None) else bmcw,
                0 if (bmch is None) else bmch
            )
            if ((rr - rl) < 1 or (rr - rl) > ims_w - 1 or (rb - rt) < 1 or
                    (rb - rt) > ims_h - 1):
                raise EimageDataError('Invalid rect borders')
            
            sw = (rr - rl)
            sh = (rb - rt)
            lm = int(sw * lmcw + sh * lmch)
            rm = int(sw * rmcw + sh * rmch)
            tm = int(sw * tmcw + sh * tmch)
            bm = int(sw * bmcw + sh * bmch)
            bw = sw + lm + rm
            bh = sh + tm + bm
            kb = bw * 1.0 / bh
            if bw < 2 or bh < 2:
                raise EimageDataError('Invalid estimated result sizes')
            
            adapt = False
            if faw or fah:
                adapt = True
                if faw == 0:
                    faw = bw
                if fah == 0:
                    fah = bh
                fw = faw
                fh = fah
            if ((fw and 
                (fw < 2 or fw > ims_w - 2)) or 
                (fh and
                (fh < 2 or fh > ims_h - 2))):
                raise EimageDataError('Invalid forced result sizes')
            
            kr = None
            if frw:
                kr = frw * 1.0 / sw
            if frh:
                kr = frh * 1.0 / sh
            if frw and frh:
                kr = max(frw * 1.0 / sw, frh * 1.0 / sh)
            
            if fw and fh:
                kf = fw * 1.0 / fh
                (mw, mh) = ((lm + rm), (tm + bm))
                if not kb == kf:
                    if bw >= bh:
                        lm = (int(round(lm * (1 - (bw - bh * kf) * 1.0 / mw)))
                              if lm > 0 else 0)
                        rm = (int(round(rm * (1 - (bw - bh * kf) * 1.0 / mw)))
                              if rm > 0 else 0)
                    else:
                        tm = (int(round(tm * (1 - (bh - bw * 1.0 / kf) * 1.0 / mh)))
                              if tm > 0 else 0)
                        bm = (int(round(bm * (1 - (bh - bw * 1.0 / kf) * 1.0 / mh)))
                              if bm > 0 else 0)
            
            bw = sw + lm + rm
            bh = sh + tm + bm
            kb = bw * 1.0 / bh
            (mw, mh) = ((lm + rm), (tm + bm))
            (kaw, kah) = ((ims_w - 1) * 1.0 / bw, (ims_h - 1) * 1.0 / bh)
            if min(kaw, kah) < 1:
                lm = int(round(lm * (1 - (bw * (1.0 - kaw)) / mw))) if lm > 0 else 0
                rm = int(round(rm * (1 - (bw * (1.0 - kaw)) / mw))) if rm > 0 else 0
                tm = int(round(tm * (1 - (bh * (1.0 - kah)) / mh))) if tm > 0 else 0
                bm = int(round(bm * (1 - (bh * (1.0 - kah)) / mh))) if bm > 0 else 0
            bw = sw + lm + rm
            bh = sh + tm + bm
            kb = bw * 1.0 / bh
            (cl, ct, cr, cb) = (
                rl - lm,
                rt - tm,
                rr + rm,
                rb + bm
            )
            xo = yo = 0
            if adapt:
                i = 0
                while any([cl < 0, ct < 0, cr > ims_w - 1, cb > ims_h - 1]) and i < self.ADOPT_MAX_STEP:
                    i += 1
                    if cl < 0:
                        cr = cr - cl - 1
                        cl = 0
                    if ct < 0:
                        cb = cb - ct - 1
                        ct = 0
                    if cr > ims_w - 1:
                        cl = cl - cr + ims_w
                        cr = ims_w - 1
                    if cb > ims_h - 1:
                        ct = ct - cb + ims_h
                        cb = ims_h - 1
                if i >= self.ADOPT_MAX_STEP:
                    raise EimageError('Margin adaptor maximum iteratioons exceed')
            else:
                if cl < 0:
                    xo = -cl
                    cl = 0
                if ct < 0:
                    yo = -ct
                    ct = 0
                if cr > ims_w - 1:
                    cr = ims_w - 1
                if cb > ims_h - 1:
                    cb = ims_h - 1

            if nodata:
                (imp_w, imp_h) = (cr-cl, cb-ct)
                imp = None
            else:    
                imp = ims.crop((cl, ct, cr, cb))
                (imp_w, imp_h) = imp.size
        
            if xo == 0 and yo == 0 and imp_w == bw and imp_h == bh:
                imd = imp
                (imd_w, imd_h) = (imp_w, imp_h)
            else:
                if nodata:
                    imd = None
                    (imd_w, imd_h) = (bw, bh)
                else:    
                    imd = Image.new("RGB", (bw, bh), (0, 0, 0))
                    imd.paste(imp, (xo, yo))
                    (imd_w, imd_h) = imd.size
        else: # nocrop
            imd = ims
            xo = yo = 0
            (cl,ct,cr,cb) = (0,0,ims_w-1,ims_h-1)    
            (imd_w, imd_h) = (ims_w, ims_h)
        
        if not noresize:
            (ww, wh) = (imd_w, imd_h)
            if fw and not fh:
                wh = int(round(wh * fw * 1.0 / ww))
                ww = fw
            if fh and not fw:
                ww = int(round(ww * fh * 1.0 / wh))
                wh = fh
            if fw and fh:
                (ww, wh) = (fw, fh)
            if kr:
                kw = max(ww * 1.0 / imd_w, wh * 1.0 / imd_h)
                if kr > kw:
                    ww = int(round(ww * (1 + kr - kw)))
                    wh = int(round(wh * (1 + kr - kw)))
            
            if not nodata:
                if ww != imd_w or wh != imd_h:
                    if wh > self.RESIZE_ALGO_MAX_HEIGHT:
                        imd = imd.resize((ww, wh), Image.NEAREST)
                    else:
                        imd = imd.resize((ww, wh), Image.BICUBIC)
                    #(imd_w, imd_h) = imd.size
        
        fr = [cl / (ims_w * 1.0), ct / (ims_h * 1.0), cr / (ims_w * 1.0),
              cb / (ims_h * 1.0)]
        tr = {"a": ((ims_w * 1.0) / (imd_w * 1.0)), "b": 0.0,
              "c": 0.0, "d": ((ims_h * 1.0) / (imd_h * 1.0)),
              "x": (xo - cl) / (imd_w * 1.0), "y": (yo - ct) / (imd_h * 1.0)}
        return {"data": imd, "transform": tr, "outerframe": fr,
                "info": self.iinfo(imd)}

    def subtitle(self, ims, i_conf, i_data, i_rects, i_transform, i_rectmargin):
        """
            conf: subtitle config and template

            data: titledata structure (key-value to substitute template fields)

            rects: red boxes to draw

            transform: transform matrix to adjust box coords

            rectmargin: margin for boxes

        """
        (ims_w, ims_h) = ims.size
        transform = i_transform if i_transform else dict(self.DEFAULT_TITLE_TRANSFORM)
        data = i_data
        conf  = dict(self.DEFAULT_TITLE_CONF)
        try:
            if self.titlefont: conf['font']['name'] = self.titlefont 
            if 'float' in i_conf: conf['float'] = i_conf['float']
            if 'height' in i_conf: conf['height'] = i_conf['height']
            if 'grid' in i_conf: conf['grid'] = i_conf['grid']
            if 'align' in i_conf: conf['align'] = i_conf['align']
            if 'padding' in i_conf: conf['padding'] = i_conf['padding']
            if 'template' in i_conf: conf['template'] = i_conf['template']
            if 'font' in i_conf: 
                if 'name' in i_conf['font']: conf['font']['name'] =  i_conf['font']['name']
                if 'size' in i_conf['font']: conf['font']['size'] =  i_conf['font']['size']
            if not isinstance(conf, dict): raise EimageDataError('Subtitle: invalid conf')
            if not "template" in conf or not isinstance(conf["template"], list): raise EimageDataError('Subtitle: invalid template')
            if not isinstance(data, dict): raise EimageDataError('Subtitle: invalid title data')
            if not isinstance(transform, dict): raise EimageDataError('Subtitle: invalid transform matrix')
            if not isinstance(i_rects, list): raise EimageDataError('Subtitle: invalid rect array')
            if not isinstance(i_rectmargin, list): raise EimageDataError('Subtitle: invalid rect margin array')
        except Exception as ex:
            print(("Subtitle config error: %s"%(str(conf))))
            raise
        try:
            font = ImageFont.truetype(conf['font']['name'],
                                    conf['font']['size'], encoding='unic')
        except Exception as ex:
            raise EimageError('Error registering subtitle font')
        (fw, fh) = font.getsize('0987654321')
        if conf['height'] == 0:
            conf['height'] = fh * 2 * conf['grid'][0]
        (dx, dy) = (int(ims_w // conf['grid'][1]),
                    int(conf['height'] // conf['grid'][0] // 2))
        imt = Image.new("RGB", (ims_w, conf['height']), (0, 0, 0))
        draw = ImageDraw.Draw(imt)
        for tpl in conf["template"]:
            try:
                token = (eval("data" + "".join([("['" + x + "']") for x in (tpl[4].replace('[', ' ').
                                                    replace(']', ' ').split())]))
                         if len(tpl[4]) > 0 else '')
            except:
                token = ''
            txt = tpl[2] + ' ' + token + tpl[3]
            (x, y) = (tpl[1] * dx + conf['padding'][0],
                      ((tpl[0] * 2 + 1) * dy - int(fh // 2)))
            draw.text((x, y), txt.decode('utf-8'), font=font)
        imr = Image.new("RGB", (ims_w, ims_h + conf['height']), (0, 0, 0))
        (imr_w, imr_h) = imr.size
        draw = ImageDraw.Draw(ims)
        for rect in i_rects:
            rect = [int(x) if x > 1.0 else int(round(x)) for x in (self.transform(rect, transform, (ims_w, ims_h)))]
            draw.rectangle([int(x) if x > 1.0 else int(round(x)) for x in ([a * (100. + b) / 100. for a, b in
                                  zip(list(rect), i_rectmargin)])], outline='red')
        if conf['float'] == 'bottom':
            imr.paste(ims, (0, 0))
            imr.paste(imt, (0, ims_h))
        else:
            imr.paste(imt, (0, 0))
            imr.paste(ims, (0, conf['height'] + 1))
        tr = {"a": ((ims_w * 1.0) / (imr_w * 1.0)), "b": 0.0, "c": 0.0,
              "d": ((ims_h * 1.0) / (imr_h * 1.0)), "x": 0.0, "y": 0.0}
        return {"data": imr, "transform": tr, "info": self.iinfo(imr)}

    """
        
        img:        PIL Image or JPEG byte string according to ImageIN
        rect:        float[4] ltrb relative coord
        preset:        preset name
        titledata:    dict with subst for subtitle template
        subrects:     float[4][] red boxes to draw
        lang:        language to select locatized subtitle config, 
                    if none - the default self.lang priority list uses
        nodata:        calculations without image processing            

    """
    def make(self, img, rect, preset, titledata, subrects,
        lang, ImageIN = ImageAsJPEG, ImageOUT = ImageAsJPEG, nodata = False):
        if not preset in self.presets:
            raise EimageDataError('No such preset')
        _preset = self.presets[preset]
        _subconf = None
        if 'subtitle' in _preset:
            if 'subconf' in _preset['subtitle']:
                if not _preset['subtitle']['subconf'] in self.subtitles:
                    raise EimageDataError('Unknown subtitle config name: %s' % (_preset['subtitle']['subconf']))
                _lang = self.lang if not lang else [lang]    
                for l in _lang:
                    if l in self.subtitles[_preset['subtitle']['subconf']]: 
                        _subconf = self.subtitles[_preset['subtitle']['subconf']][l]
                        break
                if not _subconf:
                    raise EimageDataError('Unknown language %s for subtitle config %s' % (str(_lang),_preset['subtitle']['subconf']))
            elif 'conf' in _preset['subtitle']:
                _subconf = _preset['subtitle']['conf']
            else:
                raise EimageDataError('No subtitle config defined for preset %s' % (preset))
        return self.prepare(img, rect, _preset, _subconf, titledata, subrects, ImageIN, ImageOUT, nodata)

    """
        
        img:        PIL Image or JPEG byte string according to ImageIN
        rect:        float[4] ltrb relative coord
        conf:        preset dict with:
            - sizes:    int[4-6]
            - margin:     float[6-8]
            - subtitle: 
                - subconf:        tempalte name
                - conf:            template alternate dict
                - rectmargin:    float[4] margin for red boxes
            - digest:    list ["md5","sha1","sha256"] 
        subconf:    subtitle template dict
        titledata:    dict with subst for subtitle template
        subrects:     float[4][] red boxes to draw
        nodata:     calculations without image processing

    """
    def prepare(self, img, rect, conf, subconf, titledata, subrects,
        ImageIN = ImageAsJPEG, ImageOUT = ImageAsJPEG, nodata = False):
        if ImageIN == self.ImageAsJPEG:
            img = Image.open(io.StringIO(img)).convert("RGB")
        result = self.recrop(img,rect,
            conf['margin'] if 'margin' in conf else None,
            conf['sizes'] if 'sizes' in conf else None,
            nodata)
        if subconf and result['data']:
            subresult = self.subtitle(result['data'], 
                            subconf, 
                            titledata, 
                            subrects,
                            result['transform'],
                            conf['subtitle']['rectmargin'] 
                                if 'rectmargin' in conf['subtitle'] 
                                else [0,0,0,0])    
            subresult['outerframe'] = result['outerframe']
            result = subresult
        if ImageOUT == self.ImageAsJPEG and result['data']:
            io = io.StringIO()
            io.seek(0)
            io.truncate(0)
            result['data'].save(io, "JPEG")
            result['data'] = io.getvalue()            
            if 'digest' in conf:
                result['digest'] = self.ihash(result['data'],conf['digest'])
        return result    

